//
//  AtomClient.m
//  Atom
//
//  Created by Jonathan Younger on Wed May 05 2004.
//  Copyright (c) 2004 Daikini Software.
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.

//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.

//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#import "AtomClient.h"
#import "AtomExtensions.h"
#import "AtomElement.h"


// Atom Notifications
NSString *AtomClientDidCreateAtomEntryNotification = @"AtomClient Created AtomEntry";
NSString *AtomClientDidUpdateAtomEntryNotification = @"AtomClient Updated AtomEntry";
NSString *AtomClientDidDeleteAtomEntryNotification = @"AtomClient Deleted AtomEntry";
NSString *AtomClientDidRetrieveAtomEntryNotification = @"AtomClient Retrieved AtomEntry";
NSString *AtomClientDidRetrieveAtomFeedNotification = @"AtomClient Retrieved AtomFeed";
NSString *AtomClientRequestFailedNotification = @"AtomClient Request Failed";


@implementation AtomClient (AtomClientPrivate)
- (NSString *)makeNonce
{
	NSDate *timestamp = [NSDate date];
	NSString *nonce = [[NSString stringWithFormat:@"%@:%@", [timestamp descriptionWithCalendarFormat:@"%Y-%m-%dT%H:%M:%SZ" timeZone:nil locale:nil], @"private"] stringUsingSHA1HexadecimalHash];
	return [NSString stringWithFormat:@"%@ %@", [timestamp descriptionWithCalendarFormat:@"%Y-%m-%dT%H:%M:%SZ" timeZone:nil locale:nil], nonce];
}

- (void)connection:(NSURLConnection *)aConnection didReceiveData:(NSData *)aData
{
	[[self data] appendData:aData];
}

- (void)connection:(NSURLConnection *)aConnection didFailWithError:(NSError *)anError
{			
	if ([[self delegate] respondsToSelector:@selector(client:didFailWithError:)])
	{
		[[self delegate] client:self didFailWithError:anError];
	}
	
	[[NSNotificationCenter defaultCenter] postNotificationName:AtomClientRequestFailedNotification object:nil];
	
	[entry release];
	[aConnection autorelease];
}

- (void)connectionDidFinishLoading:(NSURLConnection *)aConnection
{
	AtomEntry *atomEntry;
	AtomFeed *atomFeed;
	NSError *error;
	
	switch ([self currentOperation]) {
		case AtomClientCreatingAtomEntry:
			if ([(NSHTTPURLResponse *)[self response] statusCode] != 201) {
				error = [self errorFromResponse:[self response] data:[self data]];
				if ([[self delegate] respondsToSelector:@selector(client:didFailWithError:)]) {
					[[self delegate] client:self didFailWithError:error];
				}
				
				[[NSNotificationCenter defaultCenter] postNotificationName:AtomClientRequestFailedNotification object:self userInfo:[NSDictionary dictionaryWithObject:error forKey:@"Error"]];
			} else {
				atomEntry = [AtomEntry atomEntryFromData:[self data]];
				if ([[self delegate] respondsToSelector:@selector(client:didCreateAtomEntry:)]) {
					[[self delegate] client:self didCreateAtomEntry:atomEntry];
				}
				
				[[NSNotificationCenter defaultCenter] postNotificationName:AtomClientDidCreateAtomEntryNotification object:self userInfo:[NSDictionary dictionaryWithObject:atomEntry forKey:@"AtomEntry"]];
			}
			break;
		case AtomClientUpdatingAtomEntry:
			if ([(NSHTTPURLResponse *)[self response] statusCode] != 200) {
				error = [self errorFromResponse:[self response] data:[self data]];
				if ([[self delegate] respondsToSelector:@selector(client:didFailWithError:)]) {
					[[self delegate] client:self didFailWithError:error];
				}
				
				[[NSNotificationCenter defaultCenter] postNotificationName:AtomClientRequestFailedNotification object:self userInfo:[NSDictionary dictionaryWithObject:error forKey:@"Error"]];
			} else {
				if ([[self delegate] respondsToSelector:@selector(client:didUpdateAtomEntry:)]) {
					[[self delegate] client:self didUpdateAtomEntry:[self entry]];
				}
				
				[[NSNotificationCenter defaultCenter] postNotificationName:AtomClientDidUpdateAtomEntryNotification object:self userInfo:[NSDictionary dictionaryWithObject:[self entry] forKey:@"AtomEntry"]];
			}
			break;
		case AtomClientDeletingAtomEntry:
			if ([(NSHTTPURLResponse *)[self response] statusCode] != 200) {
				error = [self errorFromResponse:[self response] data:[self data]];
				if ([[self delegate] respondsToSelector:@selector(client:didFailWithError:)]) {
					[[self delegate] client:self didFailWithError:error];
				}
				
				[[NSNotificationCenter defaultCenter] postNotificationName:AtomClientRequestFailedNotification object:self userInfo:[NSDictionary dictionaryWithObject:error forKey:@"Error"]];
			} else {
				if ([[self delegate] respondsToSelector:@selector(client:didDeleteAtomEntry:)]) {
					[[self delegate] client:self didDeleteAtomEntry:[self entry]];
				}
				
				[[NSNotificationCenter defaultCenter] postNotificationName:AtomClientDidDeleteAtomEntryNotification object:self userInfo:[NSDictionary dictionaryWithObject:[self entry] forKey:@"AtomEntry"]];
			}
			break;
		case AtomClientRetrievingAtomEntry:
			if ([(NSHTTPURLResponse *)[self response] statusCode] != 200) {
				error = [self errorFromResponse:[self response] data:[self data]];
				if ([[self delegate] respondsToSelector:@selector(client:didFailWithError:)]) {
					[[self delegate] client:self didFailWithError:error];
				}
				
				[[NSNotificationCenter defaultCenter] postNotificationName:AtomClientRequestFailedNotification object:self userInfo:[NSDictionary dictionaryWithObject:error forKey:@"Error"]];
			} else {
				atomEntry = [AtomEntry atomEntryFromData:[self data]];
				if ([[self delegate] respondsToSelector:@selector(client:didRetrieveAtomEntry:)]) {
					[[self delegate] client:self didRetrieveAtomEntry:atomEntry];
				}
				
				[[NSNotificationCenter defaultCenter] postNotificationName:AtomClientDidRetrieveAtomEntryNotification object:self userInfo:[NSDictionary dictionaryWithObject:atomEntry forKey:@"AtomEntry"]];
			}
			break;
		case AtomClientRetrievingAtomFeed:
			if ([(NSHTTPURLResponse *)[self response] statusCode] != 200) {
				error = [self errorFromResponse:[self response] data:[self data]];
				if ([[self delegate] respondsToSelector:@selector(client:didFailWithError:)]) {
					[[self delegate] client:self didFailWithError:error];
				}
				
				[[NSNotificationCenter defaultCenter] postNotificationName:AtomClientRequestFailedNotification object:self userInfo:[NSDictionary dictionaryWithObject:error forKey:@"Error"]];
			} else {
				atomFeed = [AtomFeed atomFeedFromData:[self data]];
				if ([[self delegate] respondsToSelector:@selector(client:didRetrieveAtomFeed:)]) {
					[[self delegate] client:self didRetrieveAtomFeed:atomFeed];
				}
				
				[[NSNotificationCenter defaultCenter] postNotificationName:AtomClientDidRetrieveAtomFeedNotification object:self userInfo:[NSDictionary dictionaryWithObject:atomFeed forKey:@"AtomFeed"]];
			}
			break;
	}
	
	[self setCurrentOperation:AtomClientNoOperation];
	[entry release];
	[aConnection autorelease];
}

- (void)connection:(NSURLConnection *)aConnection didReceiveResponse:(NSURLResponse *)aResponse
{
	[self setResponse:aResponse];
}

- (NSError *)errorFromResponse:(NSURLResponse *)aResponse data:(NSData *)aData
{
	NSError *error;
	if ([self useSoap]) {
		CFXMLTreeRef tree = CFXMLTreeCreateFromData(kCFAllocatorDefault, (CFDataRef)aData,
													NULL, kCFXMLParserSkipWhitespace, kCFXMLNodeCurrentVersion);
		
		CFXMLTreeRef faultcode = [AtomElement getTreeFromParent:[AtomElement getTreeFromParent:[AtomElement getTreeFromParent:[AtomElement getTreeFromParent:tree name:@"soap:Envelope"] name:@"soap:Body"] name:@"soap:Fault"] name:@"faultcode"];
		CFXMLTreeRef faultstring = [AtomElement getTreeFromParent:[AtomElement getTreeFromParent:[AtomElement getTreeFromParent:[AtomElement getTreeFromParent:tree name:@"soap:Envelope"] name:@"soap:Body"] name:@"soap:Fault"] name:@"faultstring"];
		error = [NSError errorWithDomain:NSURLErrorDomain code:[[AtomElement getElementFromTree:faultcode] intValue] userInfo:[NSDictionary dictionaryWithObject:[AtomElement getElementFromTree:faultstring] forKey:NSLocalizedDescriptionKey]];
		CFRelease(tree);
	} else {
		error = [NSError errorWithDomain:NSURLErrorDomain code:[(NSHTTPURLResponse *)aResponse statusCode] userInfo:[NSDictionary dictionaryWithObject:[aData stringUsingEncoding:NSUTF8StringEncoding] forKey:NSLocalizedDescriptionKey]];
	}
	
	return error;
}

- (NSData *)dataFromSoapData:(NSData *)aData
{
	if (aData != nil) {
		if ([self useSoap]) {
			CFXMLTreeRef tree = CFXMLTreeCreateFromData(kCFAllocatorDefault, (CFDataRef)aData,
														NULL, kCFXMLParserSkipWhitespace, kCFXMLNodeCurrentVersion);
			CFXMLTreeRef soapTree = [AtomElement getTreeFromParent:[AtomElement getTreeFromParent:tree name:@"soap:Envelope"] name:@"soap:Body"];
			
			CFRelease(tree);
			
			CFXMLTreeRef child;
			NSString *childName;
			int index;
			for (index = CFTreeGetChildCount(soapTree) - 1; index >= 0; index--) {
				child = CFTreeGetChildAtIndex(soapTree, index);
				childName = (NSString *)CFXMLNodeGetString(CFXMLTreeGetNode(child));
				
				if ([childName isEqualToString:@"feed"] || [childName isEqualToString:@"entry"]) {
					CFDataRef xmlData = CFXMLTreeCreateXMLData(kCFAllocatorDefault, child);
					return [(NSData *)xmlData autorelease]; 
				}
			}
			
		} else {
			return aData;
		}
	}
	
	return nil;
}

- (void)sendRequest:(NSURLRequest *)aRequest delegate:(id)aDelegate
{
	connection = [[NSURLConnection alloc] initWithRequest:aRequest delegate:self];
	
	[self setDelegate:aDelegate];
	
	if (connection != nil) {
		data = [[NSMutableData alloc] init];
	} else {
		if ([[self delegate] respondsToSelector:@selector(connection:didFailWithError:)]) {
			[[self delegate] connection:connection didFailWithError:nil];
		}
	}
}

- (NSData *)sendSynchronousRequest:(NSURLRequest *)aRequest returningResponse:(NSURLResponse **)aResponse error:(NSError **)anError
{
	NSData *responseData = [[[NSURLConnection sendSynchronousRequest:aRequest
												   returningResponse:aResponse 
															   error:anError] retain] autorelease];
	
	//NSLog(@"NSURLResponse = %@ and NSError = %@", response, error);
	if ([self useSoap]) {
		return [self dataFromSoapData:responseData];
	} else {
		return responseData;
	}
}

- (NSURLRequest *)requestWithURL:(NSURL *)aURL method:(NSString *)aMethod object:(id)anObject
{
	NSData *requestData;
	NSString *contentType;
	NSMutableURLRequest *request = [[NSMutableURLRequest alloc] initWithURL:aURL];
	
	NSString *nonce = [self makeNonce];
	NSString *encodedNonce = [nonce stringUsingBase64Encoding];
	NSString *created = [[NSDate date] descriptionWithCalendarFormat:@"%Y-%m-%dT%H:%M:%SZ" timeZone:nil locale:nil];
	NSString *passwordDigest = [[[NSString stringWithFormat:@"%@%@%@", nonce, created, [self password]] stringUsingSHA1RawHash] stringUsingBase64Encoding];
	
	[request addValue:[aURL host] forHTTPHeaderField:@"Host"];
	[request addValue:[self userAgent] forHTTPHeaderField:@"User-Agent"];
	
	if ([self useSoap]) {
		[request addValue:[NSString stringWithFormat:@"http://schemas.xmlsoap.org/wsdl/http/%@", aMethod] forHTTPHeaderField:@"SOAPAction"];
		
		NSMutableString *soapRequest = [NSMutableString stringWithString:@"<soap:Envelope"];
		[soapRequest appendString:@" xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\""];
		[soapRequest appendString:@" xmlns:wsu=\"http://schemas.xmlsoap.org/ws/2002/07/utility\""];
		[soapRequest appendString:@" xmlns:wsse=\"http://schemas.xmlsoap.org/ws/2002/07/secext\">"];
		[soapRequest appendString:@"<soap:Header>"];
		[soapRequest appendString:@"<wsse:Security>"];
		[soapRequest appendString:@"<wsse:UsernameToken>"];
		[soapRequest appendFormat:@"<wsse:Username>%@</wsse:Username>", [self userName]];
		[soapRequest appendFormat:@"<wsse:Password Type=\"wsse:PasswordDigest\">%@</wsse:Password>", passwordDigest];
		[soapRequest appendFormat:@"<wsse:Nonce>%@</wsse:Nonce>", encodedNonce];
		[soapRequest appendFormat:@"<wsu:Created>%@</wsu:Created>", created];
		[soapRequest appendString:@"</wsse:UsernameToken>"];
		[soapRequest appendString:@"</wsse:Security>"];
		[soapRequest appendString:@"</soap:Header>"];
		[soapRequest appendString:@"<soap:Body>"];
		[soapRequest appendFormat:@"<%@ xmlns=\"http://schemas.xmlsoap.org/wsdl/http/\">", aMethod];
		
		if (anObject != nil) {
			[soapRequest appendString:[anObject stringValue]];
		}
		
		[soapRequest appendFormat:@"</%@>", aMethod];
		[soapRequest appendString:@"</soap:Body>"];
		[soapRequest appendString:@"</soap:Envelope>"];
		
		requestData = [soapRequest dataUsingEncoding:NSUTF8StringEncoding];
		aMethod = @"POST";
		contentType = @"text/xml";
	} else {
		[request addValue:@"WSSE profile=\"UsernameToken\"" forHTTPHeaderField:@"Authorization"];
		[request addValue:[NSString stringWithFormat:@"UsernameToken Username=\"%@\", PasswordDigest=\"%@\", Created=\"%@\", Nonce=\"%@\"", [self userName], passwordDigest, created, encodedNonce] forHTTPHeaderField:@"X-WSSE"];
		contentType = @"application/x.atom+xml";
		
		if (anObject != nil) {
			requestData = [[anObject stringValue] dataUsingEncoding:NSUTF8StringEncoding];
		} else {
			requestData = [NSData data];
		}
	}
	
	if ([request valueForHTTPHeaderField: @"Content-Type"] == nil) {
		[request addValue:contentType forHTTPHeaderField:@"Content-Type"];
	} else {
		[request setValue:contentType forHTTPHeaderField:@"Content-Type"];
	}
	
	if ([request valueForHTTPHeaderField: @"Content-Length"] == nil) {
		[request addValue:[NSString stringWithFormat:@"%d", [requestData length]] forHTTPHeaderField: @"Content-Length"];
	} else {
		[request setValue:[NSString stringWithFormat:@"%d", [requestData length]] forHTTPHeaderField: @"Content-Length"];
	}
	
	[request setHTTPMethod:aMethod];
	[request setHTTPBody: requestData];
	[request addValue:@"close" forHTTPHeaderField:@"Connection"];
	
	return (NSURLRequest *)request;
}



- (NSURLConnection *)connection { return connection; }

- (void)setConnection:(NSURLConnection *)aConnection
{
    if (connection != aConnection) {
        [aConnection retain];
        [connection release];
        connection = aConnection;
    }
}


- (NSMutableData *)data { return data; }

- (void)setData:(NSMutableData *)aData
{
    if (data != aData) {
        [aData retain];
        [data release];
        data = aData;
    }
}


- (NSURLResponse *)response { return response; }

- (void)setResponse:(NSURLResponse *)aResponse
{
    if (response != aResponse) {
        [aResponse retain];
        [response release];
        response = aResponse;
    }
}


- (id)delegate { return delegate; }

- (void)setDelegate:(id)aDelegate
{
    if (delegate != aDelegate) {
        [aDelegate retain];
        [delegate release];
        delegate = aDelegate;
    }
}


- (int)currentOperation { return currentOperation; }
- (void)setCurrentOperation:(int)aCurrentOperation
{
	currentOperation = aCurrentOperation;
}


- (AtomEntry *)entry { return entry; }

- (void)setEntry:(AtomEntry *)anEntry
{
    if (entry != anEntry) {
        [anEntry retain];
        [entry release];
        entry = anEntry;
    }
}
@end

@implementation AtomClient
- (id)initWithUserName:(NSString *)aUserName password:(NSString *)aPassword
{
	if (self = [super init]) {
		[self setUserName:aUserName];
		[self setPassword:aPassword];
		[self setUserAgent:@"Atom.framework/0.5"];
	}
	
	return self;
}

- (void)createAtomEntry:(AtomEntry *)anAtomEntry URL:(NSURL *)aURL delegate:(id)aDelegate
{
	[self setCurrentOperation:AtomClientCreatingAtomEntry];
	[self sendRequest:[self requestWithURL:aURL method:@"POST" object:anAtomEntry] delegate:aDelegate];
}

- (void)updateAtomEntry:(AtomEntry *)anAtomEntry delegate:(id)aDelegate
{
	[self setCurrentOperation:AtomClientUpdatingAtomEntry];
	// NEED TO USE SOAP WHEN DOING UPDATES BECAUSE THERE IS A BUG IN NSMutableURLRequest/NSURLConnection WHERE THE HTTP BODY IS NOT BEING SENT.
	[self setUseSoap:YES];
	[self sendRequest:[self requestWithURL:[[anAtomEntry atomLinkWithRel:@"service.edit" type:@"application/x.atom+xml"] HREF] method:@"PUT" object:anAtomEntry] delegate:aDelegate];
}

- (void)deleteAtomEntry:(AtomEntry *)anAtomEntry delegate:(id)aDelegate
{
	[self setCurrentOperation:AtomClientDeletingAtomEntry];
	[self sendRequest:[self requestWithURL:[[anAtomEntry atomLinkWithRel:@"service.edit" type:@"application/x.atom+xml"] HREF] method:@"DELETE" object:anAtomEntry] delegate:aDelegate];
}

- (void)retrieveAtomEntryWithContentsOfURL:(NSURL *)aURL delegate:(id)aDelegate
{
	[self setCurrentOperation:AtomClientRetrievingAtomEntry];
	[self sendRequest:[self requestWithURL:aURL method:@"GET" object:nil] delegate:aDelegate];
}

- (void)retrieveAtomFeedWithContentsOfURL:(NSURL *)aURL delegate:(id)aDelegate
{
	[self setCurrentOperation:AtomClientRetrievingAtomFeed];
	[self sendRequest:[self requestWithURL:aURL method:@"GET" object:nil] delegate:aDelegate];
}

- (AtomEntry *)createAtomEntry:(AtomEntry *)anAtomEntry URL:(NSURL *)aURL returningError:(NSError **)anError
{
	NSURLResponse *synchronousResponse;
	NSError *requestError;
	NSError *atomError;
	
	NSData *responseData = [self sendSynchronousRequest:[self requestWithURL:aURL method:@"POST" object:anAtomEntry] returningResponse:&synchronousResponse error:&requestError];
	
	if (responseData != nil) {
		if ([(NSHTTPURLResponse *)synchronousResponse statusCode] != 201) {
			atomError = [self errorFromResponse:synchronousResponse data:responseData];
			*anError = atomError;
		} else {
			return [AtomEntry atomEntryFromData:responseData];
		}
	} else {
		*anError = requestError;
	}
	
	return nil;
}

- (AtomEntry *)updateAtomEntry:(AtomEntry *)anAtomEntry returningError:(NSError **)anError
{
	NSURLResponse *synchronousResponse;
	NSError *requestError;
	NSError *atomError;
	
	// NEED TO USE SOAP WHEN DOING UPDATES BECAUSE THERE IS A BUG IN NSMutableURLRequest/NSURLConnection WHERE THE HTTP BODY IS NOT BEING SENT.
	[self setUseSoap:YES];
	
	NSData *responseData = [self sendSynchronousRequest:[self requestWithURL:[[anAtomEntry atomLinkWithRel:@"service.edit" type:@"application/x.atom+xml"] HREF] method:@"PUT" object:anAtomEntry] returningResponse:&synchronousResponse error:&requestError];
	
	if (responseData != nil) {
		if ([(NSHTTPURLResponse *)synchronousResponse statusCode] != 200) {
			atomError = [self errorFromResponse:synchronousResponse data:responseData];
			*anError = atomError;
		} else {
			return [AtomEntry atomEntryFromData:responseData];
		}
	} else {
		*anError = requestError;
	}
	
	return nil;
}

- (BOOL)deleteAtomEntry:(AtomEntry *)anAtomEntry returningError:(NSError **)anError
{
	NSURLResponse *synchronousResponse;
	NSError *requestError;
	NSError *atomError;
	
	NSData *responseData = [self sendSynchronousRequest:[self requestWithURL:[[anAtomEntry atomLinkWithRel:@"service.edit" type:@"application/x.atom+xml"] HREF] method:@"DELETE" object:anAtomEntry] returningResponse:&synchronousResponse error:&requestError];
	
	if (responseData != nil) {
		if ([(NSHTTPURLResponse *)synchronousResponse statusCode] != 200) {
			atomError = [self errorFromResponse:synchronousResponse data:responseData];
			*anError = atomError;
		} else {
			return YES;
		}
	} else {
		*anError = requestError;
	}
	
	return NO;
}


- (AtomEntry *)retrieveAtomEntryWithContentsOfURL:(NSURL *)aURL returningError:(NSError **)anError
{
	NSURLResponse *synchronousResponse;
	NSError *requestError;
	NSError *atomError;
	
	NSData *responseData = [self sendSynchronousRequest:[self requestWithURL:aURL method:@"GET" object:nil] returningResponse:&synchronousResponse error:&requestError];

	if (responseData != nil) {
		if ([(NSHTTPURLResponse *)synchronousResponse statusCode] != 200) {
			atomError = [self errorFromResponse:synchronousResponse data:responseData];
			*anError = atomError;
		} else {
			return [AtomEntry atomEntryFromData:responseData];
		}
	} else {
		*anError = requestError;
	}

	return nil;
}

- (AtomFeed *)retrieveAtomFeedWithContentsOfURL:(NSURL *)aURL returningError:(NSError **)anError
{
	NSURLResponse *synchronousResponse;
	NSError *requestError;
	NSError *atomError;
	
	NSData *responseData = [self sendSynchronousRequest:[self requestWithURL:aURL method:@"GET" object:nil] returningResponse:&synchronousResponse error:&requestError];
	
	if (responseData != nil) {
		if ([(NSHTTPURLResponse *)synchronousResponse statusCode] != 200) {
			atomError = [self errorFromResponse:synchronousResponse data:responseData];
			*anError = atomError;
		} else {
			return [AtomFeed atomFeedFromData:responseData];
		}
	} else {
		*anError = requestError;
	}
	
	return nil;
}


- (void)cancel
{
	[[self connection] cancel];
	[connection autorelease];
}



- (NSString *)userName { return userName; }

- (void)setUserName:(NSString *)anUserName
{
    if (userName != anUserName) {
        [anUserName retain];
        [userName release];
        userName = anUserName;
    }
}


- (NSString *)password { return password; }

- (void)setPassword:(NSString *)aPassword
{
    if (password != aPassword) {
        [aPassword retain];
        [password release];
        password = aPassword;
    }
}


- (BOOL)useSoap { return useSoap; }
- (void)setUseSoap:(BOOL)flag
{
	useSoap = flag;
}


- (NSString *)userAgent { return userAgent; }

- (void)setUserAgent:(NSString *)anUserAgent
{
    if (userAgent != anUserAgent) {
        [anUserAgent retain];
        [userAgent release];
        userAgent = anUserAgent;
    }
}
@end
